#!/usr/bin/env python
# encoding: utf-8

# This file is part of Vexy
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# SPDX-License-Identifier: Apache-2.0
# Copyright (c) Paul Horton. All Rights Reserved.

import enum


class EcoSystemType:

    def __init__(self, *, name: str, purl_type: str, description: str) -> None:
        self._name = name
        self._purl_type = purl_type
        self._description = description

    @property
    def name(self) -> str:
        return self._name

    @property
    def purl_type(self) -> str:
        return self._purl_type

    @property
    def description(self) -> str:
        return self._description


_ALL_ECOSYSTEMS = {
    'CARGO': EcoSystemType(
        name='Cargo', purl_type='cargo', description='The Rust community\'s crate registry'
    ),
    'COCOAPODS': EcoSystemType(
        name='Cocoapods', purl_type='cocoapods', description='The Cocoa dependency manager'
    ),
    'COMPOSER': EcoSystemType(
        name='Composer', purl_type='composer', description='Dependency Manager for PHP'
    ),
    'CONAN': EcoSystemType(
        name='Conan', purl_type='conan', description='The open-source C/C++ package manager'
    ),
    'CONDA': EcoSystemType(
        name='Conda', purl_type='conda', description='Conda is a cross-platform, language-agnostic binary package '
                                                     'manager'
    ),
    'CRAN': EcoSystemType(
        name='Cran', purl_type='cran', description='Comprehensive R Archive Network'
    ),
    'GO': EcoSystemType(
        name='Go', purl_type='golang', description='Go Package Managers'
    ),
    'MAVEN': EcoSystemType(
        name='Maven', purl_type='maven', description='Apache Maven'
    ),
    'NPM': EcoSystemType(
        name='NPM', purl_type='npm', description='Package manager for the JavaScript programming language'
    ),
    'NUGET': EcoSystemType(
        name='NuGet', purl_type='nuget', description='Microsoft NuGet'
    ),
    'PYPI': EcoSystemType(
        name='PyPi', purl_type='pypi', description='Python Package Index'
    ),
    'RPM': EcoSystemType(
        name='RPM', purl_type='rpm', description='Redhat Package Manager'
    ),
    'RUBY_GEM': EcoSystemType(
        name='RubyGems', purl_type='gem', description='Ruby package system'
    ),
    'SWIFT': EcoSystemType(
        name='Swift', purl_type='swift', description='Swift Package Manager'
    )
}


class EcoSystem(enum.Enum):
    """
    Languages/ecosystems to the PURL type

    Starting list taken from https://github.com/package-url/purl-spec/blob/master/PURL-TYPES.rst
    """

    BITBUCKET = 'BITBUCKET'
    CARGO = 'CARGO'
    COCOAPODS = 'COCOAPODS'
    COMPOSER = 'COMPOSER'
    CONAN = 'CONAN'
    CONDA = 'CONDA'
    CRAN = 'CRAN'
    DART = 'PUB'
    DEBIAN = 'DEB'
    DOCKER = 'DOCKER'
    FLUTTER = 'PUB'
    GENERIC = 'GENERIC'
    GITHUB = 'GITHUB'
    GO = 'GOLANG'
    HASKELL = 'HACKAGE'
    HEX = 'HEX'
    MAVEN = 'MAVEN'
    NPM = 'NPM'
    NUGET = 'NUGET'
    OCI = 'OCI'
    PYPI = 'PYPI'
    RPM = 'RPM'
    RUBY_GEM = 'GEM'
    SWIFT = 'SWIFT'

    def get_info(self) -> EcoSystemType:
        return _ALL_ECOSYSTEMS[self.value]



#!/usr/bin/env python
# encoding: utf-8

# This file is part of Vexy
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# SPDX-License-Identifier: Apache-2.0
# Copyright (c) Paul Horton. All Rights Reserved.

from decimal import Decimal
from typing import Any, Dict, Set

from cyclonedx.model import XsUri
from cyclonedx.model.component import Component
from cyclonedx.model.impact_analysis import ImpactAnalysisAffectedStatus
from cyclonedx.model.vulnerability import (
    BomTarget,
    BomTargetVersionRange,
    Vulnerability,
    VulnerabilityAdvisory,
    VulnerabilityRating,
    VulnerabilityReference,
    VulnerabilityScoreSource,
    VulnerabilitySeverity,
    VulnerabilitySource,
)
from ossindex.ossindex import OssIndex

from .. import EcoSystem
from .base import BaseSource


class OssIndexSource(BaseSource):

    def _component_complete_for_source(self, component: Component) -> bool:
        return component.purl is not None

    def _configure_source(self, config: Dict[str, Any]) -> None:
        pass

    def get_vulnerabilities(self) -> Set[Vulnerability]:
        ossi = OssIndex(enable_cache=False)
        ossi_results = ossi.get_component_report(
            packages=list(map(lambda c: c.purl, self.valid_components))
        )

        vulnerabilities: Set[Vulnerability] = set()

        for ossi_c in ossi_results:
            if ossi_c.vulnerabilities:
                for ossi_v in ossi_c.vulnerabilities:
                    v_source = VulnerabilitySource(
                        name=OssIndexSource.source_name(), url=XsUri(uri=ossi_v.reference)
                    )
                    v = Vulnerability(
                        source=OssIndexSource.source(),
                        references=[
                            VulnerabilityReference(id=ossi_v.id, source=v_source)
                        ],
                        cwes=[int(ossi_v.cwe[4:])] if ossi_v.cwe else None,
                        description=ossi_v.title,
                        detail=ossi_v.description,
                        affects=[
                            BomTarget(
                                ref=ossi_c.get_package_url().to_string(),
                                versions=[
                                    BomTargetVersionRange(
                                        version=ossi_c.get_package_url().version,
                                        status=ImpactAnalysisAffectedStatus.UNKNOWN
                                    )
                                ]
                            )
                        ]
                    )

                    if ossi_v.cvss_score:
                        v.ratings.add(
                            VulnerabilityRating(
                                source=v_source, score=Decimal(
                                    ossi_v.cvss_score
                                ) if ossi_v.cvss_score else None,
                                severity=VulnerabilitySeverity.get_from_cvss_scores(
                                    (ossi_v.cvss_score,)
                                ) if ossi_v.cvss_score else None,
                                method=VulnerabilityScoreSource.get_from_vector(
                                    vector=ossi_v.cvss_vector
                                ) if ossi_v.cvss_vector else None,
                                vector=ossi_v.cvss_vector)
                        )

                    for ext_ref in ossi_v.external_references:
                        v.advisories.add(VulnerabilityAdvisory(url=XsUri(uri=ext_ref)))

                    vulnerabilities.add(v)

        return vulnerabilities

    @staticmethod
    def source() -> VulnerabilitySource:
        """
        Instance that represents this data source.

        :return: VulnerabilitySource
        """
        return VulnerabilitySource(name=OssIndexSource.source_name(), url=XsUri(uri=OssIndexSource.source_url()))

    @staticmethod
    def source_name() -> str:
        return 'OSS Index by Sonatype'

    @staticmethod
    def source_description() -> str:
        return 'OSS Index is a free service used by developers to identify open source dependencies and determine if ' \
               'there are any known, publicly disclosed, vulnerabilities. OSS Index is based on vulnerability data ' \
               'derived from public sources and does not include human curated intelligence nor expert remediation ' \
               'guidance.'

    @staticmethod
    def source_ecosystems() -> Set[EcoSystem]:
        return {
            EcoSystem.CARGO, EcoSystem.COCOAPODS, EcoSystem.COMPOSER, EcoSystem.COMPOSER, EcoSystem.CONAN,
            EcoSystem.CONDA, EcoSystem.CRAN, EcoSystem.GO, EcoSystem.MAVEN, EcoSystem.NPM, EcoSystem.NUGET,
            EcoSystem.PYPI, EcoSystem.RPM, EcoSystem.RUBY_GEM, EcoSystem.SWIFT
        }

    @staticmethod
    def source_url() -> str:
        """
        Public URL for this data source

        :return: str
        """
        return 'https://ossindex.sonatype.org/'



# encoding: utf-8

# This file is part of Vexy
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# SPDX-License-Identifier: Apache-2.0
# Copyright (c) Paul Horton. All Rights Reserved.

from typing import Any, Dict, List, Set

from cyclonedx.model import OrganizationalContact, XsUri
from cyclonedx.model.component import Component
from cyclonedx.model.impact_analysis import ImpactAnalysisAffectedStatus
from cyclonedx.model.vulnerability import (
    BomTarget,
    BomTargetVersionRange,
    Vulnerability,
    VulnerabilityAdvisory,
    VulnerabilityCredits,
    VulnerabilityRating,
    VulnerabilityReference,
    VulnerabilitySource,
)
from osv.api import OsvApi
from osv.model import OsvPackage, OsvReferenceType, OsvSeverityType

from .. import EcoSystem
from .base import BaseSource


class OsvSource(BaseSource):

    def get_vulnerabilities(self) -> Set[Vulnerability]:
        osv = OsvApi()

        vulnerabilities: Set[Vulnerability] = set()

        for component in self.valid_components:
            osv_vulnerabilities = osv.query(package=OsvPackage(purl=component.purl))
            for osv_v in osv_vulnerabilities:
                affected_versions: List[BomTargetVersionRange] = []
                for affected in osv_v.affected:
                    if affected.ranges:
                        for r in affected.ranges:
                            affected_versions.append(BomTargetVersionRange(
                                range=r.as_purl_vers(), status=ImpactAnalysisAffectedStatus.AFFECTED
                            ))
                    for v in affected.versions:
                        affected_versions.append(BomTargetVersionRange(
                            version=v, status=ImpactAnalysisAffectedStatus.AFFECTED
                        ))

                ratings: List[VulnerabilityRating] = []
                for severity in osv_v.severity:
                    if severity.type_ == OsvSeverityType.CVSS_V3:
                        ratings.append(VulnerabilityRating(
                            vector=severity.score
                        ))

                advisories: List[VulnerabilityAdvisory] = []
                for ref in osv_v.references:
                    if ref.type_ == OsvReferenceType.ADVISORY:
                        advisories.append(VulnerabilityAdvisory(url=XsUri(ref.url)))

                credits_ = None
                if osv_v.credits:
                    credit = osv_v.credits.pop()
                    contact_1 = credit.contact.pop()
                    credits_ = VulnerabilityCredits(individuals=[
                        OrganizationalContact(
                            name=credit.name,
                            phone=contact_1 if '@' not in contact_1 else None,
                            email=contact_1 if '@' in contact_1 else None
                        )
                    ])

                vulnerabilities.add(
                    Vulnerability(
                        source=OsvSource.source(),
                        references=[
                            VulnerabilityReference(id=osv_v.id_, source=OsvSource.source())
                        ],
                        ratings=ratings,
                        cwes=None,
                        description=osv_v.summary,
                        detail=osv_v.details,
                        advisories=advisories,
                        created=osv_v.published,
                        published=osv_v.published,
                        updated=osv_v.modified if osv_v.modified else None,
                        credits=credits_,
                        affects=[
                            BomTarget(
                                ref=component.bom_ref.value,
                                versions=affected_versions
                            )
                        ],
                    )
                )

        return vulnerabilities

    def _component_complete_for_source(self, *, component: Component) -> bool:
        return component.purl is not None

    def _configure_source(self, *, config: Dict[str, Any]) -> None:
        pass

    @staticmethod
    def source() -> VulnerabilitySource:
        """
        Instance that represents this data source.

        :return: VulnerabilitySource
        """
        return VulnerabilitySource(name=OsvSource.source_name(), url=XsUri(uri=OsvSource.source_url()))

    @staticmethod
    def source_name() -> str:
        return 'OSV'

    @staticmethod
    def source_description() -> str:
        return 'An open, precise, and distributed approach to producing and consuming vulnerability information for ' \
               'open source.'

    @staticmethod
    def source_ecosystems() -> Set[EcoSystem]:
        return {
            EcoSystem.CARGO, EcoSystem.COMPOSER, EcoSystem.DEBIAN, EcoSystem.GO, EcoSystem.MAVEN, EcoSystem.NPM,
            EcoSystem.NUGET, EcoSystem.PYPI, EcoSystem.RPM, EcoSystem.RUBY_GEM
        }

    @staticmethod
    def source_url() -> str:
        """
        Public URL for this data source

        :return: str
        """
        return 'https://osv.dev/'



# encoding: utf-8

# This file is part of Vexy
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# SPDX-License-Identifier: Apache-2.0
# Copyright (c) Paul Horton. All Rights Reserved.

from typing import Any, Dict, Set

import requests
from cyclonedx.model import XsUri
from cyclonedx.model.component import Component
from cyclonedx.model.vulnerability import Vulnerability, VulnerabilitySource

from .. import EcoSystem
from .base import BaseSource


class OsvDbSource(BaseSource):

    def get_vulnerabilities(self) -> Set[Vulnerability]:
        vulnerabilities: Set[Vulnerability] = set()
        total = 0

        for component in self.valid_components:
            if component.purl:
                print(f'Querying OSV DB for known vulnerabilities for {component.purl.to_string()}')
                vulns = requests.post('http://127.0.0.1:8888/api/v1/query', json={
                    'purl': component.purl.to_string()
                }).json()

                total = total + len(vulns["results"])
            # for osv_v in osv_vulnerabilities:
            #     affected_versions: List[BomTargetVersionRange] = []
            #     for affected in osv_v.affected:
            #         if affected.ranges:
            #             for r in affected.ranges:
            #                 affected_versions.append(BomTargetVersionRange(
            #                     version_range=r.as_purl_vers(), status=ImpactAnalysisAffectedStatus.AFFECTED
            #                 ))
            #         for v in affected.versions:
            #             affected_versions.append(BomTargetVersionRange(
            #                 version=v, status=ImpactAnalysisAffectedStatus.AFFECTED
            #             ))
            #
            #     ratings: List[VulnerabilityRating] = []
            #     for severity in osv_v.severity:
            #         if severity.type_ == OsvSeverityType.CVSS_V3:
            #             ratings.append(VulnerabilityRating(
            #                 vector=severity.score
            #             ))
            #
            #     advisories: List[VulnerabilityAdvisory] = []
            #     for ref in osv_v.references:
            #         if ref.type_ == OsvReferenceType.ADVISORY:
            #             advisories.append(VulnerabilityAdvisory(url=XsUri(ref.url)))
            #
            #     credits_ = None
            #     if osv_v.credits:
            #         credit = osv_v.credits.pop()
            #         contact_1 = credit.contact.pop()
            #         credits_ = VulnerabilityCredits(individuals=[
            #             OrganizationalContact(
            #                 name=credit.name,
            #                 phone=contact_1 if '@' not in contact_1 else None,
            #                 email=contact_1 if '@' in contact_1 else None
            #             )
            #         ])
            #
            #     vulnerabilities.add(
            #         Vulnerability(
            #             source=OsvSource.source(),
            #             references=[
            #                 VulnerabilityReference(id=osv_v.id_, source=OsvSource.source())
            #             ],
            #             ratings=ratings,
            #             cwes=None,
            #             description=osv_v.summary,
            #             detail=osv_v.details,
            #             advisories=advisories,
            #             created=osv_v.published,
            #             published=osv_v.published,
            #             updated=osv_v.modified if osv_v.modified else None,
            #             credits=credits_,
            #             affects_targets=[
            #                 BomTarget(
            #                     ref=component.bom_ref.value,
            #                     versions=affected_versions
            #                 )
            #             ],
            #         )
            #     )

        print(f'Total Vulns: {total}')

        return vulnerabilities

    def _component_complete_for_source(self, *, component: Component) -> bool:
        return component.purl is not None

    def _configure_source(self, *, config: Dict[str, Any]) -> None:
        pass

    @staticmethod
    def source() -> VulnerabilitySource:
        """
        Instance that represents this data source.

        :return: VulnerabilitySource
        """
        return VulnerabilitySource(name=OsvDbSource.source_name(), url=XsUri(uri=OsvDbSource.source_url()))

    @staticmethod
    def source_name() -> str:
        return 'OSV DB'

    @staticmethod
    def source_description() -> str:
        return 'An open, precise, and distributed approach to producing and consuming vulnerability information for ' \
               'open source.'

    @staticmethod
    def source_ecosystems() -> Set[EcoSystem]:
        return {
            EcoSystem.DEBIAN
        }

    @staticmethod
    def source_url() -> str:
        """
        Public URL for this data source

        :return: str
        """
        return 'http://127.0.0.1:8888/'



#!/usr/bin/env python
# encoding: utf-8

# This file is part of Vexy
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# SPDX-License-Identifier: Apache-2.0
# Copyright (c) Paul Horton. All Rights Reserved.

from abc import ABC, abstractmethod
from typing import Any, Dict, Iterable, Optional, Set

from cyclonedx.model.component import Component
from cyclonedx.model.vulnerability import Vulnerability, VulnerabilitySource

from .. import EcoSystem


class BaseSource(ABC):

    def __init__(self, *, config: Optional[Dict[str, Any]] = None) -> None:
        if config:
            self._configure_source(config=config)

        self._all_components: Set[Component] = set()
        self._valid_components: Set[Component] = set()

    def process_components(self, *, components: Iterable[Component]) -> None:
        self._all_components = set(components)
        self._valid_components = set(filter(lambda c: self._component_complete_for_source(component=c), components))

    @property
    def all_components(self) -> Set[Component]:
        return self._all_components

    @property
    def valid_components(self) -> Set[Component]:
        return self._valid_components

    @abstractmethod
    def get_vulnerabilities(self) -> Set[Vulnerability]:
        pass

    @abstractmethod
    def _component_complete_for_source(self, *, component: Component) -> bool:
        """
        Whether the given Component has enough data (the right fields) for us to query this data source for known
        vulnerabilities.

        :param component: Component
        :return: bool
        """
        pass

    @abstractmethod
    def _configure_source(self, *, config: Dict[str, Any]) -> None:
        """
        Perform any source specific configuration such as authentication.

        :param config: Dict[str, Any]
        :return: None
        """
        pass

    @staticmethod
    @abstractmethod
    def source() -> VulnerabilitySource:
        """
        Instance that represents this data source.

        :return: VulnerabilitySource
        """
        pass

    @staticmethod
    @abstractmethod
    def source_name() -> str:
        """
        Human-friendly name for this data source.

        :return: str
        """
        pass

    @staticmethod
    @abstractmethod
    def source_description() -> str:
        """
        Human-friendly description of this data source.

        :return: str
        """
        pass

    @staticmethod
    @abstractmethod
    def source_ecosystems() -> Set[EcoSystem]:
        """
        Which ecosystems this source has vulnerability data for.

        :return: Set[str]
        """
        pass

    @staticmethod
    @abstractmethod
    def source_url() -> str:
        """
        Public URL for this data source

        :return: str
        """
        pass



#!/usr/bin/env python
# encoding: utf-8

# This file is part of Vexy
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# SPDX-License-Identifier: Apache-2.0
# Copyright (c) Paul Horton. All Rights Reserved.

import argparse
import enum
import json
import sys
from datetime import datetime
from importlib import import_module
from io import TextIOWrapper
from os import getcwd, path
from string import printable
from typing import Dict, Optional, Set, cast
from urllib.parse import quote
from xml.etree import ElementTree

import yaml
from cyclonedx.exception import CycloneDxException
from cyclonedx.model import ExternalReference, ExternalReferenceType, Tool, XsUri
from cyclonedx.model.bom import Bom
from cyclonedx.output import BaseOutput
from cyclonedx.schema import OutputFormat, SchemaVersion
from rich.console import Console
from rich.progress import Progress

from vexy.sources import ALL_SOURCES
from vexy.sources.base import BaseSource


@enum.unique
class _CLI_OUTPUT_FORMAT(enum.Enum):
    XML = 'xml'
    JSON = 'json'


_output_formats: Dict[_CLI_OUTPUT_FORMAT, OutputFormat] = {
    _CLI_OUTPUT_FORMAT.XML: OutputFormat('Xml'),
    _CLI_OUTPUT_FORMAT.JSON: OutputFormat('Json'),
}
_output_default_filenames = {
    _CLI_OUTPUT_FORMAT.XML: 'cyclonedx-vex.xml',
    _CLI_OUTPUT_FORMAT.JSON: 'cyclonedx-vex.json',
}

if sys.version_info >= (3, 8):
    from importlib.metadata import version as meta_version
else:
    from importlib_metadata import version as meta_version

try:
    __ThisToolVersion: Optional[str] = str(meta_version('vexy'))  # type: ignore[no-untyped-call]
except Exception:
    __ThisToolVersion = None
ThisTool = Tool(vendor='Vexy', name='vexy', version=__ThisToolVersion or 'UNKNOWN')
ThisTool.external_references.update([
    ExternalReference(
        type=ExternalReferenceType.BUILD_SYSTEM,
        url=XsUri('https://github.com/madpah/vexy/actions')
    ),
    ExternalReference(
        type=ExternalReferenceType.DISTRIBUTION,
        url=XsUri('https://pypi.org/project/vexy/')
    ),
    ExternalReference(
        type=ExternalReferenceType.DOCUMENTATION,
        url=XsUri('https://vexy.readthedocs.io/')
    ),
    ExternalReference(
        type=ExternalReferenceType.ISSUE_TRACKER,
        url=XsUri('https://github.com/madpah/vexy/issues')
    ),
    ExternalReference(
        type=ExternalReferenceType.LICENSE,
        url=XsUri('https://github.com/madpah/vexy/blob/main/LICENSE')
    ),
    ExternalReference(
        type=ExternalReferenceType.RELEASE_NOTES,
        url=XsUri('https://github.com/madpah/vexy/blob/main/CHANGELOG.md')
    ),
    ExternalReference(
        type=ExternalReferenceType.VCS,
        url=XsUri('https://github.com/madpah/vexy')
    )
])


class VexyCmd:
    DEFAULT_CONFIG_FILE: str = '.vexy.config'

    # Whether debug output is enabled
    _DEBUG_ENABLED: bool = False

    # Parsed Arguments
    _arguments: argparse.Namespace

    def __init__(self, args: argparse.Namespace) -> None:
        self._arguments = args
        self._console = Console()

        if self._arguments.debug_enabled:
            self._DEBUG_ENABLED = True
            self._arguments.quiet_enabled = False
            self._debug_message('!!! DEBUG MODE ENABLED !!!')
            self._debug_message('Parsed Arguments: {}'.format(self._arguments))

        self._data_sources: Set[BaseSource] = set()
        self._attempt_source_config_load(config=self._arguments.vexy_config)

        if not self._is_quiet():
            self._console.print(
                f'Vexy is configured to use [bold cyan]{len(self._data_sources)}[/bold cyan] data sources.'
            )

    def _attempt_source_config_load(self, config: TextIOWrapper) -> None:
        # Attempts to Vexy source configuration at the locations
        with config as config_f:
            vexy_config = yaml.safe_load(config_f.read())
            for source_key, source_config in vexy_config['sources'].items():
                self._data_sources.add(ALL_SOURCES[source_key](config=source_config))

    def get_cli_output_format(self) -> _CLI_OUTPUT_FORMAT:
        return _CLI_OUTPUT_FORMAT(str(self._arguments.output_format).lower())

    def _get_output_format(self) -> OutputFormat:
        return _output_formats[self.get_cli_output_format()]

    def _is_quiet(self) -> bool:
        return bool(self._arguments.quiet_enabled)

    def execute(self) -> None:
        with Progress() as progress:
            task_parse = progress.add_task(
                'Parsing CycloneDX BOM for Components', total=100, visible=not self._is_quiet()
            )
            progress.start_task(task_id=task_parse)

            # parser: BaseParser
            input_bom: Bom
            try:
                with self._arguments.input_source as input_bom_fh:
                    if str(self._arguments.input_source.name).endswith('.json'):
                        input_bom = Bom.from_json(data=json.loads(input_bom_fh.read()))  # type: ignore[attr-defined]

                    elif str(self._arguments.input_source.name).endswith('.xml'):
                        input_bom = Bom.from_xml(  # type: ignore[attr-defined]
                            data=ElementTree.fromstring(input_bom_fh.read())
                        )
            except CycloneDxException as e:
                print(f'Failure validating input BOM: {e}')
                return

            progress.update(
                task_id=task_parse, completed=100,
                description=f'Parsed {len(input_bom.components)} Components from CycloneDX SBOM'
            )

            vex = Bom()
            vex.metadata.tools.add(ThisTool)
            data_source_tasks = {}
            for data_source in self._data_sources:
                data_source_tasks[data_source.__class__] = progress.add_task(
                    f'Consulting {data_source.source_name()} for known vulnerabilities', total=100,
                    visible=not self._is_quiet()
                )
                data_source.process_components(components=input_bom.components)
                progress.update(
                    task_id=data_source_tasks[data_source.__class__], completed=25,
                    description=f'{data_source.source_name()}: Querying for {len(data_source.valid_components)} '
                                f'Components'
                )
                vulnerabilities = data_source.get_vulnerabilities()
                progress.update(
                    task_id=data_source_tasks[data_source.__class__], completed=50,
                    description=f'{data_source.source_name()}: Processing Vulnerabilities for '
                                f'{len(data_source.valid_components)} Components'
                )

                # @todo: CALL OUT ANY COMPONENTS THAT WERE NOT QUERIED

                i: int = 1
                for v in vulnerabilities:
                    for a in v.affects:
                        a.ref = f'{input_bom.urn()}#{quote(a.ref, safe=printable)}'
                    vex.vulnerabilities.add(v)
                    progress.update(
                        task_id=data_source_tasks[data_source.__class__],
                        completed=(50 + (i / len(vulnerabilities) * 50))
                    )
                    i += 1

        output_format = self._get_output_format()
        outputter = self._get_outputter(output_format=output_format, bom=vex)

        if self._arguments.output_file == '-' or not self._arguments.output_file:
            self._debug_message('Returning SBOM to STDOUT')
            print(outputter.output_as_string())
            return

        # Check directory writable
        output_file = self._arguments.output_file
        output_filename = path.realpath(
            output_file if isinstance(output_file, str) else _output_default_filenames[self.get_cli_output_format()]
        )
        self._debug_message('Will be outputting SBOM to file at: {}'.format(output_filename))
        outputter.output_to_file(filename=output_filename, allow_overwrite=self._arguments.output_file_overwrite)

    def _get_outputter(self, output_format: OutputFormat, bom: Bom) -> BaseOutput:
        schema_version = SchemaVersion['V{}'.format(
            str(self._arguments.output_schema_version).replace('.', '_')
        )]
        try:
            module = import_module(f"cyclonedx.output.{self._arguments.output_format.lower()}")
            output_klass = getattr(module, f"{output_format.value}{schema_version.value}")
        except (ImportError, AttributeError):
            raise ValueError(f"Unknown format {output_format.value.lower()!r}") from None

        return cast(BaseOutput, output_klass(bom=bom))

    @staticmethod
    def get_arg_parser(*, prog: Optional[str] = None) -> argparse.ArgumentParser:
        arg_parser = argparse.ArgumentParser(prog=prog, description='Vexy VEX Generator')

        arg_parser.add_argument(
            '-c', '--config', action='store', type=argparse.FileType('r'),  # FileType does handle '-'
            dest='vexy_config', required=True, default=f'{getcwd()}/{VexyCmd.DEFAULT_CONFIG_FILE}',
            help='Configuration file for Vexy defining data sources to use and their configuration.'
        )
        arg_parser.add_argument('-q', action='store_true', help='Quiet - no console output', dest='quiet_enabled')
        arg_parser.add_argument('-X', action='store_true', help='Enable debug output', dest='debug_enabled')

        input_method_group = arg_parser.add_argument_group(
            title='Input CycloneDX BOM',
            description='Where Vexy shall obtain it\'s input'
        )
        input_method_group.add_argument(
            '-i', '--in-file', action='store', metavar='FILE_PATH',
            type=argparse.FileType('r'),  # FileType does handle '-'
            default=None, dest='input_source', required=True,
            help='CycloneDX BOM to read input from. Use "-" to read from STDIN.'
        )

        output_group = arg_parser.add_argument_group(
            title='VEX Output Configuration',
            description='Choose the output format and schema version'
        )
        output_group.add_argument(
            '--format', action='store',
            choices=[f.value for f in _CLI_OUTPUT_FORMAT], default=_CLI_OUTPUT_FORMAT.XML.value,
            help='The output format for your SBOM (default: %(default)s)',
            dest='output_format'
        )
        output_group.add_argument(
            '--schema-version', action='store', choices=['1.4'], default='1.4',
            help='The CycloneDX schema version for your VEX (default: %(default)s)',
            dest='output_schema_version'
        )
        output_group.add_argument(
            '-o', '--o', '--output', action='store', metavar='FILE_PATH', default=True, required=False,
            help='Output file path for your SBOM (set to \'-\' to output to STDOUT)', dest='output_file'
        )
        output_group.add_argument(
            '--force', action='store_true', dest='output_file_overwrite',
            help='If outputting to a file and the stated file already exists, it will be overwritten.'
        )

        return arg_parser

    def _debug_message(self, message: str) -> None:
        if self._DEBUG_ENABLED:
            print('[DEBUG] - {} - {}'.format(datetime.now(), message))

    @staticmethod
    def _error_and_exit(message: str, exit_code: int = 1) -> None:
        print('[ERROR] - {} - {}'.format(datetime.now(), message))
        exit(exit_code)


def main(*, prog_name: Optional[str] = None) -> None:
    parser = VexyCmd.get_arg_parser(prog=prog_name)
    args = parser.parse_args()
    VexyCmd(args).execute()


if __name__ == "__main__":
    main()
